#!/bin/bash

function save_env() {
	env > ${TESTLOG_DIR}/${TESTNAME}.debug.env 2>&1
}

function print_var() {
	local message=$1
	local content=$2
	if [ -n "$content" ]; then
		echo "$message $content"
	fi
}

function print_test_info() {
	echo "start $1 testing..."
	print_var "device name: " "$DEVICE_NAME"
	print_var "driver name: " "$DRIVER"
	print_var "module name: " "$MODULE"
	print_var "device path: " "$DEVPATH"
	echo

	if [ -n "$PCI_BUS_DEVICE_FUNCTION" ]; then
		echo "device pci info:"
		lspci -s "$PCI_BUS_DEVICE_FUNCTION"  -vvvvv
		echo
	fi
}

function test_pass() {
	save_env
	echo "$TESTNAME test PASS!"
	exit 0
}

function test_fail() {
	save_env
	echo "$TESTNAME test FAIL!"
	exit 1
}

function run_cmd() {
    local the_cmd="$1"
    echo "$the_cmd"
    echo ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"
    eval "$the_cmd"
    [ $? -ne 0 ] && echo "failed to run $the_cmd" && test_fail
    echo ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>"
    echo;echo;echo
}

function show_module_info() {
	local module=$1
	echo
	echo "module name $module, detail info:"
	if [ -n "$module" ]; then
		modinfo "$module"
	else
		echo "warning: module name is empty!"
	fi
}

function clean_up_disk_partition() {
	local dev_path=$1

	echo "clean up all partitions for $dev_path"
	local mount_folder
	mount_folder="$(mount | egrep "^$dev_path" | awk '{print $3}')"
	for folder in $mount_folder; do
		echo "umount folder $folder"
		umount "$folder"
		if [ $? -ne 0 ]; then
			local mount_pid
			mount_pid="$(lsof | grep -w "$folder" | awk '{print $2}')"
			echo "kill pid $mount_pid"
			kill -9 "$mount_pid"
			sleep 3
			umount "$folder" || { echo "fail to umount $folder"; test_fail; }
		fi
	done

	echo "wipe out device $dev_path"
	wipefs -f -a "$dev_path" || { echo "fail to wipe out $dev_path"; test_fail; }
}

function check_partition_ready() {
	partition=$1
	echo "check partition $partition status"
	sleep 1
	for ((i=0; i < 10; i++)); do
		if ls "$partition" > /dev/null 2>&1; then
			break
		fi
		sleep 1
	done
	echo "partition $partition is ready, partition info:"
	lsblk "$partition"
}

function show_network_info() {
	echo "get all the link information:"
	ifconfig -a

	echo "device path: $DEVPATH"
	local bdf
	bdf="$(echo "$DEVPATH" | awk -F'/' '{print $NF}')"

	echo "device info:"
	lspci -vvvv -s "$bdf"
}

function show_drive_info() {
	local dev_name=${1:-"${DEVNAME}"}
	lsblk "$DEVNAME"
	echo
	fdisk -l "$DEVNAME"
	echo
	mount
	echo
	df -P
}

function check_mount_point() {
	local dev_name=${1:-"${DEVNAME}"}
	echo "check mount point $dev_name"
	if mount | grep -q "$dev_name"; then
		echo "has mount point for $dev_name:"
		mount | grep "$dev_name"
		return 1
	fi
	echo "dose not have mount point for $dev_name"
	return 0
}

function create_ext4_from_raw_disk() {
	local test_dir=$1
	local dev_name=${2:-"${DEVNAME}"}
	local partition_num=${3:-"1"}
	echo "setup partition $dev_name"
	echo -e 'n\np\n\n\n+8G\nwq\np\n' | fdisk "$dev_name"
	check_partition_ready "${dev_name}${partition_num}"
	echo "create ext4 file system on $dev_name"
	mkfs.ext4 -F "${dev_name}${partition_num}"
	echo "make test dir $test_dir"
	mkdir $test_dir
	echo "mount $test_dir"
	mount -t ext4 "${dev_name}${partition_num}" $test_dir
	[ $? -eq 0 ] || { echo "failed to mount $test_dir"; test_fail; }
	check_mount_point "$test_dir"
        echo "check mount point"
	mount | grep "$test_dir"
	echo "ls command check mount dir:"
	ls -l $test_dir
	[ $? -eq 0 ] || { echo "failed to access $test_dir"; test_fail; }
	echo
}

function generate_fio_config_file() {
	local fio_config_file=$1
	local test_dir=$2

cat > "$fio_config_file" <<FIOFILE
[global]
ioengine=sync
iodepth=4
numjobs=1
group_reporting
size=128M
direct=1
runtime=120
timeout=900

[verify-write]
bs=4k
name=iops_seqwrite
filename=${test_dir}/testfio_verifyrw
offset=0x0000
verify=md5
do_verify=1
verify_pattern=0xff
verify_dump=1
verify_fatal=1
verify_backlog=1
verify_interval=4k
rw=write
stonewall

[verify-read]
bs=4k
name=iops_seqread
filename=${test_dir}/testfio_verifyrw
offset=0x0000
verify=md5
do_verify=1
verify_pattern=0xff
verify_dump=1
verify_fatal=1
verify_backlog=1
verify_interval=16k
rw=read
stonewall

[seq-write]
bs=4k
name=iops_seqwrite
filename=${test_dir}/testfio_seqwrite
rw=write
stonewall

[seq-read]
bs=4k
name=iops_seqread
filename=${test_dir}/testfio_seqread
rw=read
stonewall

[rand-read]
bs=16k
name=iops_randread
filename=${test_dir}/testfio_randread
rw=randread
stonewall

[rand-write]
bs=64k
name=iops_randwrite
filename=${test_dir}/testfio_randwrite
rw=randwrite
stonewall

[randrw]
bs=512k
name=iops_randrw
filename=${test_dir}/testfio_randrw
rwmixread=70
rw=randrw
stonewall
FIOFILE
}

function run_fio_test() {
	local test_type=${1:-"storage"}
	local fio_config_file=$TESTLOG_DIR/fio_config_file.$$
	local test_dir=/ancert.${test_type}.fio.$$

	if $IS_RAW_DISK; then
		check_mount_point "$DEVNAME"
		if [ $? -ne 0 ]; then
			echo "failed, $DEVNAME has mount point"
			test_fail
		fi
		echo "$DEVNAME is a free disk"
		if [ "$test_type" == "nvme" ]; then
			create_ext4_from_raw_disk "$test_dir" "$DEVNAME" "p1"
		else
			create_ext4_from_raw_disk "$test_dir" "$DEVNAME"
		fi
	else
		check_mount_point "$ANCERT_TEST_MOUNT_POINT"
		if [ $? -ne 1 ]; then
			echo "failed, dose not have mount point $ANCERT_TEST_MOUNT_POINT"
			test_fail
		fi
		if [ "$ANCERT_TEST_MOUNT_POINT" != "/" ]; then
			rm -rf $ANCERT_TEST_MOUNT_POINT/ancert.${test_type}.fio.*
			test_dir=$ANCERT_TEST_MOUNT_POINT/ancert.${test_type}.fio.$$
		fi
		mkdir -p $test_dir
	fi

	echo
	echo "test dir is $test_dir"
	echo
	generate_fio_config_file "$fio_config_file" "$test_dir"
	echo
	echo "change work dir to $TESTLOG_DIR"
	cd $TESTLOG_DIR
	run_cmd "fio $fio_config_file"
	cd -
	echo "current work dir `pwd`"

	if $IS_RAW_DISK; then
		umount $test_dir
		clean_up_disk_partition "$DEVNAME"
	fi	

	if [ -d $test_dir ]; then
		rm -rf $test_dir
	fi
}
